
# initialize libraries.
import numpy
import logging
from classDef import Peak
from scipy.special import psi
from scipy.stats.distributions import norm
from operator import attrgetter

root_logger = logging.getLogger("")
debug = root_logger.debug
info = root_logger.info

def get_candidate_window2(read, x, y, repx, repy, threshold):
    # using PHI = 1e6 to prescreen the genome 
    PHI = 1e6
    GAMMA_HAT = 1.0
    tau = numpy.sqrt( y*( repx*x*(PHI+y) + repy*y*(PHI+x))/repx/repy/PHI/x**3)
    gamma = y/x
    z = (numpy.log(gamma)-numpy.log(GAMMA_HAT))*gamma/tau
    pvalue = norm.cdf(-z)
    pre_idx_list = numpy.where(pvalue[10:-10]<threshold)[0]+10
    return numpy.array(pre_idx_list)


def weighted_log_likelihood(v_hat, m, n, reads, diff_test):
    '''This function returns the derivative of the log likelihood
       of current and adjacent windows using a triangular weight.'''
    equation = 0
    n_window = len(reads)
    baseline = numpy.mean(reads[n_window/2,0:m])
    for idx in range(n_window):
        x = reads[idx, 0:m]  # x/m refer the test sample
        y = reads[idx, m:(m+n)]  # y/n refer to the control sample
        weight_x = numpy.mean(x)/baseline
        weight_y = numpy.mean(y)/baseline
        if n == 1: 
            weight_y = 0
        log_likelihood = (-(m*weight_x+n*weight_y)*psi(v_hat) +
                numpy.sum(psi(v_hat+x))*weight_x + 
                numpy.sum(psi(v_hat+y))*weight_y + 
                m*numpy.log(v_hat/(v_hat+numpy.mean(x)))*weight_x + 
                n*numpy.log(v_hat/(v_hat+numpy.mean(y)))*weight_y)
        equation = (equation + 
                log_likelihood*(1-(abs(n_window/2-idx)/(n_window/2+1))))
    return equation


def estimate_area_dispersion_factor(read, m, n, idx, peaktype, diff_test):
    '''This function implements a simple method to find the root
       of the derivative of the log likelihood.'''
    BASE = 10
    N_ITERATION = 15
    if peaktype == "sharp":
        N_WINDOWS = 1 
    elif peaktype  == "broad":
        N_WINDOWS = 10
    left = -1.0
    right = 6.0
    if weighted_log_likelihood(BASE**right, m, n, 
            read[(idx-N_WINDOWS):(idx+N_WINDOWS+1)], diff_test) > 0:
        return BASE**right
    elif weighted_log_likelihood(BASE**left, m, n,
            read[(idx-N_WINDOWS):(idx+N_WINDOWS+1)], diff_test) < 0:
        return BASE**left
    for times in xrange(N_ITERATION):
        temp = (right+left)/2
        if weighted_log_likelihood(10**temp, m, n,
                read[(idx-N_WINDOWS):(idx+N_WINDOWS+1)], diff_test) > 0:
            left = temp
        else:
            right = temp
    return BASE**right


def cal_FDR(peak_list, num_tests):
    '''Calculate the Benjamini-Hochberg FDR'''
    peak_list = sorted(peak_list, key=attrgetter('pvalue'))
    #calculate BH q-values
    q_list = [item.pvalue*num_tests/(idx+1) 
            for idx, item in enumerate(peak_list)]
    q_min = q_list[-1]
    for i in xrange(len(q_list)-1, -1, -1):
        if q_list[i] >= q_min:
            q_list[i] = q_min
        else:
            q_min = q_list[i]
    for i in range(len(q_list)):
        peak_list[i] = Peak(peak_list[i].chr, peak_list[i].index,
                            peak_list[i].pvalue, q_list[i])
    peak_list = sorted(peak_list, key=attrgetter('chr', 'index'))
    return peak_list


def negative_binomial(readData, peakfilename, swap, parameter):
    '''the main function that test for significant windows.'''
    # Initialize the parameters
    peaktype = parameter.peaktype
    remove_artefacts = parameter.remove_artefacts
    narrow_peak = parameter.narrow_peak_width
    threshold = parameter.threshold
    windowsize = parameter.window_size
    # Indicate the data 
    read_dict = readData.reads_dict 
    strands_dict = readData.data_dict_by_strands
    chr_list = readData.chr_list
    if parameter.difftest is True: 
        test_list = readData.chip1_filename_list
        control_list = readData.chip2_filename_list
    else: 
        test_list = readData.chip1_filename_list
        control_list = readData.input1_filename_list
    num_tests = readData.genome_size/windowsize
    peakfile = open(peakfilename, 'w')

    #compute number of replicates
    test_rep = len(test_list)
    control_rep = len(control_list)
    start1 = 0
    end1 = start2 = test_rep
    end2 = test_rep+control_rep
    # if swap 
    if swap is True: 
        test_rep, control_rep = control_rep, test_rep
        test_list, control_list = control_list, test_list

    # initialize basic array structures
    #sig_index_dict = {}
    sig_peaks_list = []    

    for chr in chr_list:

        read_array = read_dict[chr]
        read_array[numpy.where(read_array ==0)] = 1 
        y_bar_array = numpy.mean(read_array[:, start1:end1], 1)
        x_bar_array = numpy.mean(read_array[:, start2:end2], 1)
        if swap: #swap the chip and control reads. 
            x_bar_array, y_bar_array = y_bar_array, x_bar_array
        # setting the minimum # of reads in each window to 1 
        # so that they won't have arithmetic errors. 
    
        cand_index = get_candidate_window2(read_array, x_bar_array,
                y_bar_array, control_rep, test_rep, threshold)
        debug("there are %d candidate windows for %s", len(cand_index), chr)
        if not swap: 
            disp_list = numpy.array([estimate_area_dispersion_factor(read_array, 
                    test_rep, control_rep, idx, peaktype, parameter.difftest)
                    for idx in cand_index])
        else: 
            disp_list = numpy.array([estimate_area_dispersion_factor(read_array,
                    control_rep, test_rep, idx, peaktype, parameter.difftest)
                    for idx in cand_index])
        debug("finished estimating dispersion for %s", chr)
        cand_x_bar_array = x_bar_array[cand_index]
        cand_y_bar_array = y_bar_array[cand_index]
        gamma_array = cand_y_bar_array / cand_x_bar_array
        tau_hat_array = numpy.sqrt(cand_y_bar_array* 
                ((control_rep*cand_x_bar_array*(disp_list+cand_y_bar_array)) + 
                (test_rep*cand_y_bar_array*(disp_list+cand_x_bar_array)))/
                (test_rep*control_rep*disp_list*(cand_x_bar_array**3)))

        gamma_hat = 1.0 #Null hypothesis
        z_score_array = ((numpy.log(gamma_array)-numpy.log(gamma_hat))* 
                gamma_array/tau_hat_array)
        pval_array = norm.cdf(-z_score_array)
        # record results for potential windows
#        window_out = open(chr+"candidate.out",'w')
#        for i in range(len(cand_index)):
#            window_out.write(chr+'\t'+str(cand_index[i])+'\t'+str(disp_list[i])+'\t'+str(pval_array[i])+'\n')
#        window_out.close()
        # Record the indices of the windows that have p-value smaller
        # than the threshold. 
        test_index = numpy.where(pval_array<threshold)  
        test_index = test_index[0]
        sig_index = cand_index[test_index]
        sig_pval = pval_array[test_index]
        sig_disp = disp_list[test_index]
        for i, a in enumerate(test_index):
            sig_peaks_list.append(Peak(chr, sig_index[i], sig_pval[i], 0))

    #calculate the BH FDR. 
    debug("begin estimating fdr")
    sig_peaks_list = cal_FDR(sig_peaks_list, num_tests)
    debug("finished estimating fdr")

    # merge adjacent significant peaks. 
    info ("Merging adjacent significant windows...")
    final_peak_list = []
    for chr in chr_list:
        sig_peak_list_by_chr = \
                [item for item in sig_peaks_list if item.chr==chr]
        if len(sig_peak_list_by_chr) == 0:
            continue  # if there is no significant peak in this chromosome, skip it.
        sig_index = [item.index for item in sig_peak_list_by_chr]
        sig_pval = [item.pvalue for item in sig_peak_list_by_chr]
        sig_qval = [item.qvalue for item in sig_peak_list_by_chr]     
        sig_start, sig_end, sig_pval, sig_qval = merge_sig_window(sig_index,
                sig_pval, sig_qval, peaktype)
        for idx in xrange(len(sig_start)):
            final_peak = [chr, sig_start[idx]*windowsize/2, 
                          sig_end[idx]*windowsize/2+windowsize,
                          sig_pval[idx], sig_qval[idx]]
            final_peak_list.append(final_peak) 

    # post-processing 
    if remove_artefacts is True: 
         dump_file = open(peakfilename[:-10]+"_filtered_peaks.txt","w")
         dump_file.write("chr\tstart\tend\twidth\tp-value\tq-value\t")
         dump_file.write("chip_input_ratio\ttag_monopoly\t")
         dump_file.write("strand_overlap_original\tstrand_overlap_shifted")
    info("begin post-processing...")
    for final_peak in final_peak_list: 
        chr = final_peak[0]
        start = final_peak[1]
        end = final_peak[2]
        pval = final_peak[3]
        qval = final_peak[4]

        if swap is False: 
            chip_list = readData.chip1_filename_list
            input_list = readData.input1_filename_list
        else: 
            chip_list = readData.chip2_filename_list
            input_list = readData.input2_filename_list
        if narrow_peak or remove_artefacts: 
            (start, end, chip_input_ratio, tag_monopoly, overlap_orig, 
             overlap_roll) = post_processing_per_peak(
                    strands_dict, chip_list, input_list, chr, start, end,
                    readData.shift_size, readData.read_length,
                    narrow_peak, remove_artefacts)
            if remove_artefacts and (chip_input_ratio > 0.5 or 
                    tag_monopoly > 0.5 or (overlap_orig > 0.2 and
                overlap_roll/overlap_orig < 0.5)):
                dump_file.write(chr+"\t" + str(start) + '\t' + str(end) + '\t' +
                        str(end-start) + '\t' + str(pval) + '\t' + str(qval) + 
                        '\t' + str(chip_input_ratio) + '\t' + 
                        str(tag_monopoly) + '\t' + str(overlap_orig) +
                        '\t' + str(overlap_roll) + '\n')
                continue
        peakfile.write(chr + "\t" + str(start) + '\t' + str(end) + '\t' + 
                str(end-start) + '\t' + str(pval) + '\t' + str(qval) + '\n')
    return 

def merge_sig_window(index_list, pval_list, qval_list, peaktype):
    ''' Merge significant windows that are nearby. '''
    if peaktype=="sharp":
        # Mininal number of windows required. 
        MIN_WINDOW = 1 
        # The maximal gap (max-1) of significant windows allowed to merge.
        MAX_WINDOW = 2 
    elif peaktype =="broad":
        MIN_WINDOW = 2 
        MAX_WINDOW = 5
    sig_peak_start = []
    sig_peak_end = []
    sig_pval = []
    sig_qval = []
    for idx, pos in enumerate(index_list):
        if idx == 0:
            sig_peak_start = []
            sig_peak_end = []
            start = pos
            pre_start = pos
            peak_pval_list = [pval_list[idx]]
            peak_qval_list = [qval_list[idx]]
        else: 
            if pos - pre_start <= MAX_WINDOW:
                pre_start = pos
                peak_pval_list.append(pval_list[idx])
                peak_qval_list.append(qval_list[idx])

            elif pos-pre_start > MAX_WINDOW:
                end = pre_start
                if (end-start>=MIN_WINDOW-1):
                    sig_peak_start.append(start)
                    sig_peak_end.append(pre_start)
                    sig_pval.append(min(peak_pval_list))
                    sig_qval.append(min(peak_qval_list))
                start = pos
                pre_start = pos
                peak_pval_list = [pval_list[idx]]
                peak_qval_list = [qval_list[idx]]

    if (pre_start-start>MIN_WINDOW-1):
        sig_peak_start.append(start)
        sig_peak_end.append(pre_start)
        sig_pval.append(min(peak_pval_list))
        sig_qval.append(min(peak_qval_list))

    return sig_peak_start, sig_peak_end, sig_pval, sig_qval


def post_processing_per_peak(strands_dict, chip_list, input_list, chr,
                             start, end, shiftSize, readLength, narrow_peak,
                             remove_artefacts):
    ''' Remove artefacts and refine peak width.'''
    chip_forward = numpy.zeros(end-start)
    chip_reverse = numpy.zeros(end-start)
    input_forward = numpy.zeros(end-start)
    input_reverse = numpy.zeros(end-start)

    for chip in chip_list:
        forward = strands_dict[chr][chip]['f']
        reverse = strands_dict[chr][chip]['r']
        forward_read = forward[numpy.where( (forward > start-shiftSize[chip]) &
                (forward < end-shiftSize[chip]) )]
        reverse_read = reverse[numpy.where( (reverse > start+shiftSize[chip]) &
                (reverse < end+shiftSize[chip]) )]
        for read in forward_read:
            chip_forward[(read-start)] +=1
        for read in reverse_read:
            try: chip_reverse[(read-start-readLength)] +=1
            except IndexError:
                pass
    # using input reads to remove artefacts
    if remove_artefacts is True: 
        for input in input_list:
            forward = strands_dict[chr][input]['f']
            reverse = strands_dict[chr][input]['r']
            forward_read = forward[numpy.where( (forward > 
                    start-shiftSize[chip]) & (forward < end-shiftSize[chip]) )]
            reverse_read = reverse[numpy.where( (reverse > 
                    start+shiftSize[chip]) & (reverse < end+shiftSize[chip]) )]
            for read in forward_read:
                input_forward[(read-start)] +=1
            for read in reverse_read:
                try: input_reverse[(read-start-readLength)] +=1
                except IndexError:
                    pass

        chip_both = chip_forward + chip_reverse
        input_both = input_forward + input_reverse
        if sum(chip_both) == 0:
            pass
        else:
            chip_both = chip_both/sum(chip_both)
        if sum(input_both) ==0:
            pass
        else:
            input_both = input_both/sum(input_both)
        overlap_chip_input = numpy.sum(numpy.minimum(chip_both, input_both))
        chip_both.sort()
        chip_3_maximum = numpy.sum(chip_both[-3:])
        if sum(chip_reverse) != 0:
            chip_reverse = chip_reverse/sum(chip_reverse) 
        chip_forward_roll = numpy.roll(chip_forward,readLength)
        if sum(chip_forward) != 0:
            chip_forward = chip_forward/sum(chip_forward)
            chip_forward_roll = chip_forward_roll/sum(chip_forward)
        overlap_orig = numpy.sum(numpy.minimum(chip_forward, chip_reverse))
        overlap_orig = numpy.max([overlap_orig, 1e-5])
        overlap_roll = numpy.sum(numpy.minimum(chip_forward_roll, chip_reverse))
    else: 
        overlap_chip_input = 0
        chip_3_maximum = 0 
        overlap_orig = 1e-5
        overlap_roll = 0

    if narrow_peak is True:
        sum_forward = 0
        sum_reverse = 0
        if sum(chip_forward) > 0: 
            chip_forward = chip_forward/sum(chip_forward)
            for i in xrange(end-start): 
                sum_forward += chip_forward[i]
                if sum_forward > 0.2: 
                    new_start = start + i
                    break
        if sum(chip_reverse) > 0:
            chip_reverse = chip_reverse/sum(chip_reverse)
            for i in xrange(end-start-1, -1, -1):
                sum_reverse += chip_reverse[i]
                if sum_reverse > 0.2:
                    new_end = start + i
                    break
        if sum(chip_forward) == 0:
            new_start = new_end - 2*shiftSize
        if sum(chip_reverse) == 0:
            new_end = new_start + 2*shiftSize
        start = new_start
        end = new_end

    return (start, end, overlap_chip_input, chip_3_maximum, 
            overlap_orig, overlap_roll)


def shift_size_per_peak(strands_dict, chip_list, chr, start, end,
                        shiftSize, readLength, narrow_peak):
    '''Deprecated function. Used to estimate the shift size of 
       peak'''
    shift_list = []
    start_list = []
    end_list = []

    for chip in chip_list:
        forward = strands_dict[chr][chip]['f']
        reverse = strands_dict[chr][chip]['r']

        forward_read = forward[numpy.where( (forward > start-shiftSize[chip]) &
                (forward < end-shiftSize[chip]) )]
        reverse_read = reverse[numpy.where( (reverse > start+shiftSize[chip]) &
                (reverse < end+shiftSize[chip]) )]
        if len(reverse_read) ==0: 
            fold = 100  
        # If there is no reads in the reverse strand, define 
        # the fold (forward over reverse) to be 100 to avoid 
        # arithmetic errors.
        else:
            fold = float(len(forward_read))/len(reverse_read)
        # estimate the shift size
        forward = numpy.zeros(end-start)
        reverse = numpy.zeros(end-start)
        for read in forward_read:
            forward[(read-start):(read-start+readLength)] +=1
        for read in reverse_read:
            reverse[(read-start-readLength):(read-start)] +=1
        shade_max = 0
        shift_max = 0
        # Iterate from 0 to 250 bp, find the optimum shift that produce 
        # the most overlap between the two strands. 
        for shift in xrange(-readLength, 250-readLength): 
            forward_temp = numpy.roll(forward, shift)
            shade = numpy.sum(numpy.min(numpy.vstack(
                    [forward_temp, reverse]), 0))
            if shade > shade_max:
                shift_max = shift
                shade_max = shade
        shift_list.append(shift_max+readLength)
        if narrow_peak:
            forward_read.sort()
            reverse_read.sort()
            forward_len = len(forward_read)
            reverse_len = len(reverse_read)
            if forward_len>0:
                start_list.append(forward_read[numpy.floor(forward_len*0.2)])
            if reverse_len>0: 
                end_list.append(reverse_read[-numpy.floor(reverse_len*0.2)])
    if narrow_peak:
        if (len(start_list) ==0) | (len(end_list)==0):
            return -1
        else:
            return min(start_list), max(end_list), shift_list, fold
    else:
        return start, end, shift_list, fold



#logConfig.py


import logging
import sys
import time

#configure root settings.
#logging.basicConfig(level=logging.DEBUG, format = "%(levelname)s %(name)s %(asctime)s %(message)s", datefmt = "%m/%d/%Y %I:%M:%S %p")


#Create logging format for the handlers
#format_basic = logging.Formatter("%(levelname)-10s %(name)s %(asctime)s # %(message)s", datefmt = "%m/%d/%Y %I:%M:%S")
format_basic = logging.Formatter("%(levelname)-7s %(asctime)s # %(message)s", datefmt = " %I:%M:%S")
format_log =logging.Formatter("%(message)s")

#Create handler that prints INFO messages to stderr
info_hand = logging.StreamHandler(sys.stderr)
info_hand.setLevel(logging.DEBUG)
info_hand.setFormatter(format_basic)


#Create handlers that prints DEBUG messages to log files
log_filename = time.strftime("%m-%d-%Y-%I.%M")+"-debug.log"
debug_hand = logging.FileHandler(log_filename, 'w')
debug_hand.setLevel(logging.DEBUG)
debug_hand.setFormatter(format_basic)

'''
shiftSize_hand = logging.FileHandler("shift.log", 'w')
shiftSize_hand.setLevel(logging.DEBUG)
shiftSize_hand.setFormatter(format_log)

windowSize_hand = logging.FileHandler("window.log", 'w')
windowSize_hand.setLevel(logging.DEBUG)
windowSize_hand.setFormatter(format_log)
'''


#Create a root logger
root_log = logging.getLogger("")
root_log.setLevel(logging.DEBUG)
root_log.addHandler(info_hand)
root_log.addHandler(debug_hand)

'''
#Create a logger that handles the shift sizes
shiftSizeEst_log = logging.getLogger("shiftSizeEst")
shiftSizeEst_log.addHandler(shiftSize_hand)

#Create a logger that handles the window sizes
windowSizeEst_log = logging.getLogger("windowSizeEst")
windowSizeEst_log.addHandler(windowSize_hand)
'''





#!/usr/bin/env python
''' The main .py file for the PePr pipeline '''

import re
import os
import sys
import logging

import logConfig
import optParser
import fileParser
import shiftSize
import windowSize
import sigTests
import misc
from classDef import Parameters

def main(argv):
    # initialize the logger 
    root_logger = logging.getLogger("")
    debug = root_logger.debug
    info = root_logger.info
    # performing the option parser
    opt = optParser.opt_parser(argv)
    parameter, readData = optParser.process_opt(opt)
    # 1. read and parse the data
    fileParser.parse(readData, parameter.file_format)
    # 2. remove the redundant reads
    if (parameter.remove_redundant):
        readData.remove_redundant_reads()    
    # 3. shiftSize estimation and shifting reads
    shiftSize.estimate_shift_size(readData, parameter)
    shiftSize.shift_reads(readData)
    # 4. calculating the normalization constant 
    windowSize.estimate_normalization_constant(readData, parameter)
    # 5. windowSize estimation and split reads into windows
    windowSize.estimate_window_size(readData, parameter)
    info (" The windowSize is %s", parameter.window_size)
    windowSize.separate_exact_by_window(readData, parameter) 
    # 6. calling peaks
    if parameter.difftest is False:
        swap = False
        peakfilename = parameter.name+"__PePr_peaks.bed"
        sigTests.negative_binomial(readData, peakfilename, swap, parameter)
    else: 
        up_peakfilename = parameter.name+"__PePr_up_peaks.bed"
        swap = False
        sigTests.negative_binomial(readData, up_peakfilename, swap, parameter)
        down_peakfilename = parameter.name+"__PePr_down_peaks.bed"
        swap = True
        sigTests.negative_binomial(readData, down_peakfilename,
                                   swap, parameter)
    # 7. Write to a file that record the command and parameters.     
    parameter.write_parameter_to_file() 
 
if __name__ == '__main__':
    try: main(sys.argv)
    except KeyboardInterrupt: 
        print "user interrupted me"
        
        




import logging

import misc

rootlogger = logging.getLogger("")
info = rootlogger.info
debug = rootlogger.debug


class ReadData:
    ''' A data structure for the read data'''
    def __init__(self, chip1, input1, chip2, input2, diff_test):
        self.data_dict_by_strands = {}
        self.data_dict = {}
        self.reads_dict = {}
        self.chip1_filename_list = chip1
        self.input1_filename_list = input1
        self.chip2_filename_list = chip2
        self.input2_filename_list = input2
        self.chip_filename_list = chip1[:]
        self.input_filename_list = input1[:]
        if diff_test is True: 
            self.chip_filename_list += chip2
            self.input_filename_list += input2   
        self.filename_list = self.chip_filename_list + self.input_filename_list
        self.chr_list = []
        self.chr_length_dict = {} 
        self.read_total_per_file = {}
        self.normalization_constant = {}
        self.genome_size = 0
        self.shift_size = {}
        self.read_length = 0

    def cal_genome_size(self):  # calculate the genome size.
        for chr in self.chr_list:
            self.genome_size += self.chr_length_dict[chr]
        debug("The genome size is %d", self.genome_size)

    def cal_read_total(self):  
    # calculate the total number of reads for each sample.    
        for  chr in self.chr_list:
            for filename in self.filename_list:
                for strand in ["f", "r"]:
                    try: self.read_total_per_file[filename] += \
                        len(self.data_dict_by_strands[chr][filename][strand])
                    except KeyError: self.read_total_per_file[filename] = \
                        len(self.data_dict_by_strands[chr][filename][strand])

    def remove_redundant_reads(self):
        ''' remove additional redundant reads that are not warranted
        by a binomial test.'''
        def rm_max(list, max):
        # remove reads that are more than the maximum specified. 
            list_new = []
            pre = 0
            count = 0

            for x in list:
                if x == pre: 
                    count += 1
                else:
                    pre = x
                    count = 1
                if count > max:
                    continue
                list_new.append(x)
            return list_new
                    

        for filename in self.filename_list:
            max = misc.binomial(
                     self.read_total_per_file[filename], 1.0/self.genome_size)
            # calculate maximum duplicates for each genomic location.
            info ("The maximum at one position for %s is %d", filename, max)
            total_before = self.read_total_per_file[filename]
            total_after = 0
            for chr in self.chr_list:
                for strand in ["f", "r"]:
                    reads = self.data_dict_by_strands[chr][filename][strand]
                    reads.sort()
                    reads = rm_max(reads, max)
                    total_after += len(reads)
                    self.data_dict_by_strands[chr][filename][strand] = reads    
            self.read_total_per_file[filename] = total_after
            debug("Total # of reads before removing redundancies is %d", 
                    total_before)
            debug("Total # of reads after removing redundancies is %d",
                    total_after)
            info("The percentage of redundant reads for %s is %f ",
                    filename, 1-float(total_after)/total_before)
            
    def refine_chr_list(self): 
        '''Delete the chromosomes that having 0 reads. Find the maximum 
           genomic location in each chromosome'''
        chr_list = []
        chr_length_dict = {}
        for chr in self.data_dict_by_strands:
            temp_max_chr = 0
            all_file_has_this_chr = True
            for file in self.filename_list:
                if file not in self.data_dict_by_strands[chr]:
                    all_file_has_this_chr = False
                    break
                max_read_f = max(self.data_dict_by_strands[chr][file]['f'])
                max_read_r = max(self.data_dict_by_strands[chr][file]['r'])
                temp_max_chr = max(temp_max_chr, max_read_f, max_read_r)
            if all_file_has_this_chr is True:
                chr_list.append(chr)
                chr_length_dict[chr] = temp_max_chr
        chr_list.sort()
        self.chr_list= chr_list
        self.chr_length_dict = chr_length_dict


class Parameters:
    "store the parameters that will be used in many processes." 
    # ---- input parameters ---- #
    # ---- function specific parameters ---- #

    def __init__(self, opt):
        if opt.parameter is not "":
            self.process_parameter_file(opt.parameter)
	else: 
            self.process_command_line_option(opt)
        self.validate_parameters()

    def process_parameter_file(self, parameter_file):
        # using a case/switch commands
        file = open(parameter_file, 'r')
        for line in file: 
            items = line.split("=")
            key = items[0].strip().lower()
            value = item[1].strip()
        pass 
        # ---- will implement this at a later time. ---- #
        
    def process_command_line_option(self, opt):
        self.chip1 = opt.chip1
        self.input1 = opt.input1
        self.chip2 = opt.chip2
        self.input2 = opt.input2
        self.file_format = opt.file_format.lower()
        self.shift_size = opt.shift_size
        self.window_size = opt.window_size 
        self.difftest = opt.difftest
        self.name = opt.name
        self.remove_redundant = opt.remove_redundant
        self.threshold = opt.threshold
        self.peaktype = opt.peaktype.lower()
        self.remove_artefacts = opt.remove_artefacts
        self.narrow_peak_width = opt.narrow_peak_width
        self.unsave_log = opt.unsave_log

    def validate_parameters(self):
        if self.chip1 =="":
            raise Exception("Please specify ChIP-1 sample names")
        if self.input1 =="":
            raise Exception("Please specify input-1 sample names")
        if self.difftest is True:  # also check group 2 files if difftest
            if self.chip2 =="":
                raise Exception("Please specify ChIP-2 sample names")
            if self.input2 =="":
                raise Exception("Please specify input-2 sample names")
        if self.file_format not in [
                'bed', 'eland_multi', 'eland_extended',
                'bowtie', 'sam', 'bam'
                ]:
            raise Exception('''Please specify a valid file format: bed,
            eland_multi, eland_extended, bowtie, sam, or bam''')
        if self.peaktype not in ['sharp', 'broad']: 
            raise Exception('''please specify a peak type: sharp or broad.
            Typically, sharp works for TF better and broad
            for histone modifications.''')
        shift_list = self.shift_size.split(',')
        if len(shift_list) != 1: 
            if self.difftest is True: 
                chip_file_num = len(self.chip1.strip().split(',')) + \
                                len(self.chip2.strip().split(','))
            else: 
                chip_file_num = len(self.chip1.strip().split(','))
                
            if len(shift_list) != chip_file_num:
                raise Exception ('''Number of shift sizes are not equal
                                    to the ChIP files.''')
        if self.narrow_peak_width is True and self.peaktype == "broad":
            raise Exception('''Cannot refine peak width for broad peaks. 
                               Please set peaktype to sharp.''') 

        if self.difftest is True: 
            if len(self.chip1.strip().split(',')) == \
                    len(self.input1.strip().split(',')):
                self.chip1_matched_input = True
            else: 
                self.chip1_matched_input = False
            if len(self.chip2.strip().split(',')) == \
                    len(self.input2.strip().split(',')):
                self.chip2_matched_input = True
            else: 
                self.chip2_matched_input = False

    def write_parameter_to_file(self):
        fileout = open(self.name+"__PePr_parameters.txt", 'w')
        attributes = vars(self)
        keys_list = attributes.keys()
        keys_list.sort()
        for keys in keys_list:
            fileout.write(keys+'\t'+str(attributes[keys])+'\n')
        fileout.close()


class Peak:
    "data structure that contains the significant peaks" 

    def __init__(self, chr, index, pvalue, qvalue):
        self.chr = chr
        self.index = index
        self.pvalue = pvalue
        self.qvalue = qvalue
    def __repr__(self):
        return repr((self.chr, self.index, self.pvalue, self.qvalue))




#!/usr/bin/env python

import sys
import re
import math
import random
import numpy
import numpy.lib.stride_tricks as std
from scipy.stats import rankdata
import misc
import logging

root_logger = logging.getLogger("")
debug = root_logger.debug
info = root_logger.info
error = root_logger.error

window_logger = logging.getLogger("windowSizeEst")

def get_window_size2(array, bin=20, iter=100):
    # this is the function that estimate the window size. 
    chr_len = array.size
    rowNum = chr_len/bin
    array_window = std.as_strided(array, (rowNum, bin), 
            (bin*array.itemsize, 1*array.itemsize))
    array_window = numpy.sum(array_window, 1)
    peak_len_list = []
    for x in range(iter):
        peak = numpy.max(array_window)
        peak_idx = numpy.where(array_window == peak)[0][0]
        # set the peak to -1 so that it won't confuse the next iteration.
        array_window[peak_idx]= -1 
        if peak_idx == 0:  # Check if the window reaches the left boundary.
            left_boundary_not_reached = False
        else:
            left_boundary_not_reached = True    
        # Check if the window reaches of the right boundary.
        if peak_idx == rowNum-1: 
            right_boundary_not_reached = False
        else: 
            right_boundary_not_reached = True
        i_l = 1
        i_r = 1 
        # Be careful if anything equals to -1. 
        # Though it's unlikely that it will happen.
        while left_boundary_not_reached:

            while (array_window[peak_idx-i_l] >= 0.1*peak): 
                array_window[peak_idx-i_l] = -1
                if peak_idx-i_l==0:
                    left_boundary_not_reached = False
                    break
                i_l += 1
            if left_boundary_not_reached:
                if peak_idx-i_l==0:
                    array_window[peak_idx-i_l] = -1
                    left_boundary_not_reached = False
            
            if left_boundary_not_reached: 
            # Will continue counting if the next window(one window gap) 
            # has above the 10% mode reads. 
                if (array_window[peak_idx-i_l-1] >= 0.1*peak): 
                    array_window[peak_idx-i_l] = -1
                    i_l += 1
                else:
                    break

        while right_boundary_not_reached:
            while (array_window[peak_idx+i_r] >= 0.1*peak):
                array_window[peak_idx+i_r] = -1
                if peak_idx+i_r ==rowNum-1:
                    right_boundary_not_reached = False
                    break
                i_r += 1
            if right_boundary_not_reached:
                if peak_idx+i_r==rowNum-1:
                    array_window[peak_idx+i_r] = -1
                    right_boundary_not_reached = False

            if right_boundary_not_reached:
                if (array_window[peak_idx+i_r+1] >=0.1*peak): 
                    array_window[peak_idx+i_r] = -1
                    i_r += 1
                else:
                    break
        peak_len_list.append((1+i_l+i_r)*bin)

    return misc.median(peak_len_list)

def estimate_window_size(readData, parameter):
    if (parameter.window_size != -1):
        # window size should be even number 
        if parameter.window_size % 2 ==1:
            info("warning: window size should be even number; adding 1 to it.")
            parameter.window_size = parameter.window_size + 1
        return parameter.window_size

    info( "begin window_size calculation...")
    window_size_list = []
    for chr in readData.chr_list:
        reads = []
        for chip_filename in readData.chip_filename_list:
            reads.extend(readData.data_dict[chr][chip_filename])
        coord_array = numpy.zeros(readData.chr_length_dict[chr],
                dtype=numpy.int64)
        for x in reads:
            try: coord_array[x] += 1
            except IndexError: pass #debug("coordinates out of range")

        window = get_window_size2(coord_array)
#        if window >1000:  # limit the window size to be less than 1000 bases.
#            window = 1000
        window_size_list.append(window)
        window_logger.debug("%-10s %s", chr, window)
    debug("length of windowsize list is "+str(len(window_size_list)))
    window_median = misc.median(window_size_list)
    # Window size cannot be odd number since we need 
    # to use overlapping half windows. 
    if window_median % 2 == 1:
        window_median = window_median+1
    window_logger.debug("window_size: %+10s", window_median)
    info ( "finishing window size calculation...")
    parameter.window_size = window_median
    return  


def count_reads_by_file(read_data, bin, row_num):
    array = numpy.zeros(row_num, dtype=numpy.float64)
    for x in read_data:
        try: array[x/bin] += 1
        except IndexError:    pass  # out of range
    if any(array<0):
        error("array has reads less than 0")
    array = array + numpy.roll(array, -1)
    return array


def separate_exact_by_window(readData, parameter, normalize = "Large"):
    ''' split genome into windows'''
    window_size = parameter.window_size
    data_by_window_dict = {}
    index_dict = {}
    move_size = window_size/2

    for chr in readData.chr_list:
        info( "partitioning by window on \t%s...", chr)
        chr_len =readData.chr_length_dict[chr]
        row_num = chr_len/move_size - 1
        if parameter.difftest is False: 
            for idx, filename in enumerate(readData.filename_list):
                array = count_reads_by_file(readData.data_dict[chr][filename],
                                            move_size, row_num)
                array = (
                        array*readData.normalization_constant[filename])
                if idx ==0:
                    chr_array = array
                else:
                    chr_array = numpy.column_stack((chr_array, array))   
        else: 
            for idx, filename in enumerate(readData.chip1_filename_list):
                array = count_reads_by_file(readData.data_dict[chr][filename],
                                            move_size, row_num)
                array = (
                        array*readData.normalization_constant[filename])
                if idx ==0:
                    chip1_array = array
                else:
                    chip1_array = numpy.column_stack((chip1_array, array))
            if parameter.chip1_matched_input is True: 
                for idx, filename in enumerate(readData.input1_filename_list):
                    array = count_reads_by_file(
                            readData.data_dict[chr][filename],
                            move_size, row_num)
                    array = (
                            array*readData.normalization_constant[filename])
                    if idx ==0:
                        input1_array = array
                    else:
                        input1_array = numpy.column_stack((input1_array, array))
            else:
                for idx, filename in enumerate(readData.input1_filename_list):
                    array = count_reads_by_file(
                            readData.data_dict[chr][filename],
                            move_size, row_num)
                    array = (
                            array*readData.normalization_constant[filename])
                    if idx ==0:
                        input1_array = array
                    else:
                        input1_array += array
                input1_array /= len(readData.input1_filename_list)
                input1_array.resize(row_num, 1)
            chip1_array -= input1_array
            chip1_array[chip1_array<0] = 0
            for idx, filename in enumerate(readData.chip2_filename_list):
                array = count_reads_by_file(readData.data_dict[chr][filename],
                                            move_size, row_num)
                array = (
                        array*readData.normalization_constant[filename])
                if idx ==0:
                    chip2_array = array
                else:
                    chip2_array = numpy.column_stack((chip2_array, array))
            if parameter.chip2_matched_input is True:
                for idx, filename in enumerate(readData.input2_filename_list):
                    array = count_reads_by_file(readData.data_dict[chr][filename],
                                                move_size, row_num)
                    array = (
                            array*readData.normalization_constant[filename])
                    if idx ==0:
                        input2_array = array
                    else:
                        input2_array = numpy.column_stack((input2_array, array))
            else:
                for idx, filename in enumerate(readData.input2_filename_list):
                    array = count_reads_by_file(readData.data_dict[chr][filename],
                                                move_size, row_num)
                    array = (
                            array*readData.normalization_constant[filename])
                    if idx ==0:
                        input2_array = array
                    else:
                        input2_array += array
                input2_array /= len(readData.input2_filename_list)
                input2_array.resize(row_num, 1)
            chip2_array -= input2_array
            chip2_array[chip2_array<0] = 0
            chr_array = numpy.column_stack((chip1_array, chip2_array))
        data_by_window_dict[chr] = chr_array
    readData.reads_dict = data_by_window_dict
#    for chr in readData.chr_list:
#        fileopen = open(chr+".reads.window", 'w')
#        for row in data_by_window_dict[chr]:
#            for reads in row: 
#                fileopen.write(str(reads)+'\t')
#            fileopen.write('\n')
#        fileopen.close()
    del readData.data_dict # this is not useful anymore
    return


def estimate_normalization_constant(readData, parameter): 
    '''Estimate the normalization constant for all samples'''
    # Split the genome into 1kb windows.
    bin = 1000
    array_dict = {}
    for file in readData.filename_list: 
        array = numpy.array([], dtype=numpy.float64)
        for chr in readData.chr_list:
            row_num = readData.chr_length_dict[chr]/bin
            array_by_chr = numpy.zeros(row_num, dtype=numpy.float64)
            for x in readData.data_dict[chr][file]:
                try: array_by_chr[x/bin] += 1
                except IndexError: pass 
            array = numpy.append(array, array_by_chr)
        array_dict[file] = array
    # Create a mixed chip sample and use it as the reference 
    mixed_chip_array = numpy.array([], dtype=numpy.float64)
    for idx, chip in enumerate(readData.chip_filename_list): 
        if idx == 0: 
            mixed_chip_array = array_dict[chip].copy()
            rep_rank_sum = rankdata(-array_dict[chip])
        else: 
            mixed_chip_array += array_dict[chip]
            rep_rank_sum += rankdata(-array_dict[chip])
    mixed_chip_array /= len(readData.chip_filename_list)
    # Estimate the input normalization constant using NCIS 
    for input in readData.input_filename_list: 
        norm_constant = estimate_input_normalization(
                mixed_chip_array, array_dict[input])
        readData.normalization_constant[input] = norm_constant
        debug("The scaling factor for %s is %s", input, norm_constant)
    # Estiamte the chip normalization cosntant using modified TMM method
    for chip in readData.chip_filename_list: 
        norm_constant = estimate_chip_normalization(
                mixed_chip_array, array_dict[chip], rep_rank_sum)
        readData.normalization_constant[chip] = norm_constant
        debug("The scaling factor for %s is %s", chip, norm_constant)
    return
def estimate_input_normalization(ref, target):
    '''Estimate the input normalization constant using the NCIS method'''
    combined = ref + target 
    READ_MAX = 200
    MIN_GENOME_COVERAGE = 0.75
    pre_ratio = 1
    for r_cut in xrange(1, READ_MAX):
        index = numpy.where(combined <= r_cut)[0]
        percent_genome_covered = float(len(index))/len(combined)
        ratio = numpy.sum(ref[index])/numpy.sum(target[index])
        if percent_genome_covered > MIN_GENOME_COVERAGE and ratio > pre_ratio:
            return ratio
        else: 
            pre_ratio = ratio
    return ratio

def estimate_chip_normalization(ref, target, rep_rank_sum): 
    '''Estimate the ChIP normalization constant against the mixed ChIP
       using the modified TMM method'''
    N_PEAKS_GRID = numpy.array([1000, 5000, 10000, 20000, 30000, 40000, 50000])
    TRIM_M = 0.2
    TRIM_A = 0.05
    order = numpy.argsort(rep_rank_sum)
#    order = numpy.argsort(-ref)
    len_target_not_zero = len(numpy.where(target > 0)[0])
    ref = ref[order]
    target = target[order]
    tmm_array = numpy.array([])
    for n in N_PEAKS_GRID: 
        if n > len_target_not_zero: 
            break
        ref_n = ref[range(n)]
        ref_n[ref_n==0] = 1
        target_n = target[range(n)]
        target_n[target_n==0] = 1
        Mg = numpy.log2(ref_n/target_n)
        Ag = 0.5*numpy.log2(ref_n*target_n)
        Mg_sorted = Mg.copy()
        Mg_sorted.sort()
        Mg_lower_bound = Mg_sorted[n*TRIM_M]
        Mg_upper_bound = Mg_sorted[n*(1-TRIM_M)]
        Ag_sorted = Ag.copy()
        Ag_sorted.sort()
        Ag_lower_bound = Ag_sorted[n*TRIM_A]
        Ag_upper_bound = Ag_sorted[n*(1-TRIM_A)]
        trim_index = numpy.where((Mg > Mg_lower_bound) &
            (Mg < Mg_upper_bound) & (Ag > Ag_lower_bound) &
            (Ag < Ag_upper_bound))
        ref_trim = ref_n[trim_index]
        target_trim = target_n[trim_index]
        Mgk = numpy.log2(ref_trim/target_trim)
        Wgk = 0.5*numpy.log2(ref_trim*target_trim)
        tmm = 2**(numpy.sum(Mgk*Wgk)/numpy.sum(Wgk))
        tmm_array = numpy.append(tmm_array, tmm)
        debug("The TMM estiamted from top %s windows is %s", n, tmm)
    library_ratio = numpy.sum(ref)/numpy.sum(target)
    tmm_diff_array = numpy.abs(tmm_array - library_ratio)
    tmm_max = tmm_array[numpy.argmax(tmm_diff_array)]
    return tmm_max



import sys
import misc
import logging
import logConfig
import numpy


root_logger =  logging.getLogger("")
info = root_logger.info
debug = root_logger.debug
warning = root_logger.warning
shift_logger = logging.getLogger("shiftSizeEst")

def shift_size_per_chrom(forward, reverse, file=-1):
    #estimate the shift size for each chromosome separately. 

    shift_max = 200
    
    max_overlapping =  len(list(set(forward) & set(reverse)))
    optimum_shift = 0
    
    # iterate from 0 offset to shift_max, to find which offset produce the maximal overlap.  
    for offset in xrange(1, shift_max):
        
        overlapping = set([x+offset for x in forward]) & set([y-offset for y in reverse])
        overlapping = len(list(overlapping))

        if overlapping > max_overlapping:
            max_overlapping = overlapping
            optimum_shift = offset
        if file !=-1:
            file.write( str(offset)+'\t'+str(overlapping) + '\n')
    return optimum_shift

def estimate_shift_size(readData, parameter):
    # If shift size provided by the user, then skip estimating it.
    if parameter.shift_size != "-1": 
        shift_list = parameter.shift_size.split(',')
        if len(shift_list)== 1:
            for name in readData.filename_list: 
                readData.shift_size[name] = int(shift_list[0])
                info("%-10s %s", name, shift_list[0])
        else: 
            if parameter.difftest is True: 
                for idx, name in enumerate(readData.chip_filename_list):
                    readData.shift_size[name] = int(shift_list[idx])
                    info("%-10s %s", name, shift_list[idx])
                chip1_shift_list = [readData.shift_size[chip]
                                    for chip in readData.chip1_filename_list]
                input1_shift_size = sum(chip1_shift_list)/len(chip1_shift_list)
                for name in readData.input1_filename_list:
                    readData.shift_size[name] = input1_shift_size
                chip2_shift_list = [readData.shift_size[chip]
                                    for chip in readData.chip2_filename_list]
                input2_shift_size = sum(chip2_shift_list)/len(chip2_shift_list)
                for name in readData.input2_filename_list:
                    readData.shift_size[name] = input2_shift_size
            else: 
                for idx, chip in enumerate(readData.chip1_filename_list):
                    readData.shift_size[chip] = int(shift_list[idx])
                    info("%-10s %s", chip, shift_list[idx])
                for input in readData.input1_filename_list:
                    readData.shift_size[input] = \
                            sum([int(x) for x in shift_list])/len(shift_list)
                    info("%-10s %s", input, 
                          sum([int(x) for x in shift_list])/len(shift_list))
        return 
    info("begin estimating the shift size...")
    for chip_filename in readData.chip1_filename_list:
        info("estimating for %-10s", chip_filename)
        shift_list = []
        for count, chr in enumerate(readData.chr_list):
            # estimate the shift size for five chromosomes and take
            # the median of these as the estimator. 
            if count ==5:   
                break
            forward = readData.data_dict_by_strands[chr][chip_filename]['f']
            reverse = readData.data_dict_by_strands[chr][chip_filename]['r']
            shift = shift_size_per_chrom(forward, reverse)
            info("%-10s %d", chr, shift)
            shift_list.append(shift)
        shift_median = misc.median(shift_list[:])
        info("%-10s %d", chip_filename, shift_median)
        readData.shift_size[chip_filename] = shift_median
    chip1_shift_list = [readData.shift_size[chip] 
                        for chip in readData.chip1_filename_list]
    input1_shift_size = sum(chip1_shift_list)/len(chip1_shift_list)
    for input_filename in readData.input1_filename_list: 
        readData.shift_size[input_filename] = input1_shift_size
            
    if parameter.difftest is True:  # If we're calling differential binding. 
        for filename in readData.chip2_filename_list: 
            info("estimating for %-10s", filename)
            shift_list = []
            for count, chr in enumerate(readData.chr_list):
                if count==5: 
                    break
                forward = readData.data_dict_by_strands[chr][filename]['f']
                reverse = readData.data_dict_by_strands[chr][filename]['r']
                shift = shift_size_per_chrom(forward, reverse)
                info("%-10s %d", chr, shift)
                shift_list.append(shift)
            shift_median = misc.median(shift_list[:])
            info("%-10s %d", filename, shift_median)
            readData.shift_size[filename] = shift_median
        chip2_shift_list = [readData.shift_size[chip]
                            for chip in readData.chip2_filename_list]
        input2_shift_size = sum(chip2_shift_list)/len(chip2_shift_list)
        for input_filename in readData.input2_filename_list: 
            readData.shift_size[input_filename] = input2_shift_size
    return 

    
def shift_reads(readData): #shift the reads according the specific shift size estimated or provided by user
    readData.data_dict = {}
    for chr in readData.chr_list:
        debug("shifting the reads for %s...", chr)
        readData.data_dict[chr] ={}
        for file in readData.filename_list:
            forward=readData.data_dict_by_strands[chr][file]['f']
            reverse=readData.data_dict_by_strands[chr][file]['r']
            readData.data_dict_by_strands[chr][file]['f'] = numpy.array(readData.data_dict_by_strands[chr][file]['f'])
            readData.data_dict_by_strands[chr][file]['r'] = numpy.array(readData.data_dict_by_strands[chr][file]['r'])
            readData.data_dict[chr][file]=[x+readData.shift_size[file] for x in forward]+[y-readData.shift_size[file] for y in reverse]

#    return data_dict            




from distutils.core import setup

setup(name="PePr",
      version="1.0.3b",
      description="Peak-calling and Prioritization pipeline for replicated ChIP-Seq data",
      author="Yanxiao Zhang",
      author_email="troublezhang@gmail.com",
      url="https://code.google.com/p/pepr-chip-seq/",
      license="GNU GPL v3",
      classifiers=[
          "Development Status :: 4 - Beta", 
          "Intended Audience :: Developers",
          "Intended Audience :: Science/Research",
          "Topic :: Scientific/Engineering :: Bio-Informatics",
          "Programming Language :: Python",
          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
          ],
      packages=["PePr"],
      package_data={"PePr":['data/*.bed']},
      install_requires=[
          'numpy',
          'scipy',
          'pysam',
          ]
      )






from optparse import OptionParser
from classDef import ReadData
from classDef import Parameters


def opt_parser(argv):
    parser = OptionParser()
    parser.add_option(
            "-p", "--parameter-file", action="store", type="string",
            dest="parameter", default="", 
            help="provide a file that contain the parameters",
            metavar="PARAMETER-FILE")
    parser.add_option(
            "-c", "--chip1", action="store", type="string", dest="chip1",
            default="", help="chip1 file names separated by comma",
            metavar="CHIP1")
    parser.add_option(
            "-i", "--input1", action="store",
            type="string", dest="input1", default="",
            help="input1 file names separated by comma",
            metavar="INPUT1")
    parser.add_option(
            "--chip2", action="store", type="string", dest="chip2",
            default="", help="chip2 file names separated by comma",
            metavar="CHIP2")
    parser.add_option(
            "--input2", action="store", type="string", dest="input2",
            default="", help="input2 file names separated by comma",
            metavar="INPUT2")
    parser.add_option(
            "-f", "--file-format", action="store", type="string", 
            dest="file_format",
            help="bed, sam, bam, eland_multi, eland_extended, bowtie...",
            metavar="FORMAT")
    parser.add_option(
            "-s", "--shiftsize", action="store",
            type="string", dest="shift_size", default="-1",
            help="Half the fragment size.", metavar="SHIFTSIZE")
    parser.add_option(
            "-w", "--windowsize", action="store",
            type="int", dest="window_size", default=-1,
            help="Window sizes",
            metavar="WINDOWSIZE")
    parser.add_option(
            "--diff", action="store_true",
            dest = "difftest", default=False,
            help="Perform differential binding instead of peak-calling")
    parser.add_option(
            "-n", "--name", action = "store",
            type="string", dest="name", default = "NA",
            help = "the experimental name. NA if none provided",
            metavar="NAME")
    parser.add_option(
            "-r", "--remove_duplicate", action ="store_true",
            dest = "remove_redundant", default=False,
            help="Remove duplicated reads")
    parser.add_option(
            "--threshold", action ="store",
            type='float', dest="threshold", default=1e-5,
            help="p-value threshold. Default 1e-5.")
    parser.add_option(
            "--peaktype", action="store",
            type="string", dest="peaktype", default="broad",
            help="sharp or broad. Default broad.")
    parser.add_option(
            "--remove_artefacts", action="store_true",
            dest="remove_artefacts", default=False,
            help = 'Remove PCR-duplication peaks in post-procesing')
    parser.add_option(
            "--narrow_peak_width", action="store_true",
            dest ="narrow_peak_width", default=False,
            help = '''Narrow peak width to contain the most 
            enriched regions. Only available for SHARP peak type''')
    parser.add_option(
            "--no_log", action="store_true",
            dest = "unsave_log", default=False,
            help = "Disable saving the log files")
    (opt, args)=parser.parse_args(argv)

    return opt

def process_opt(opt):
    ''' validate the parameters that the user specified'''
    parameter = Parameters(opt)

    ## parse the chip and input filename 
    chip1_filename_list = parameter.chip1.strip().split(',')
    input1_filename_list = parameter.input1.strip().split(',')
    chip2_filename_list = parameter.chip2.strip().split(',')
    input2_filename_list = parameter.input2.strip().split(',')
    ## initialize the data structure
    read_data = ReadData(
        chip1_filename_list, input1_filename_list,
        chip2_filename_list, input2_filename_list,
        parameter.difftest 
        )
    #add shift size validations
    return parameter, read_data
